
------------------------------------------------------------------------
-- Filename: EX.vhd
-- Title: Implementation of the execution stage of the MIPS pipeline 
-- Author: Yvan Ton-That
-- Date: February 2009
------------------------------------------------------------------------


library ieee;
use	ieee.std_logic_1164.all;
use	ieee.std_logic_arith.all;




entity EX_STAGE is
port (

	-- Control Inputs
	ALUOp_p	  	: in std_logic;	-- the ALU performs an operation based on the opcode field of the instruction if '0' and based on the funct field of the instruction if '1'	
	ALUSrcB_p		: in std_logic; -- select the contents of register B if '0' and the sign extended immediate if '1'
	
	-- Datapath Inputs
	Instr_p		: in std_logic_vector(31 downto 0);
	Imm_p 		: in std_logic_vector(31 downto 0);	
	A_p		: in std_logic_vector(63 downto 0);
	B_p		: in std_logic_vector(63 downto 0);
	
	-- Outputs
	ALUOuput_p 		: out std_logic_vector(63 downto 0)
	
);
end EX_STAGE;




architecture implementation of EX_STAGE is


-- ALU input and output signals
signal ALUin1_s, ALUin2_s, ALUout_s 				: std_logic_vector(63 downto 0);

-- Other signals
signal opcode_s					: std_logic_vector(5 downto 0);
signal funct_s					: std_logic_vector(5 downto 0);
signal signextend_s					: std_logic_vector(31 downto 0);


begin


ALUOuput_p <= ALUout_s;

opcode_s <= Instr_p(31 downto 26);
funct_s <= Instr_p(5 downto 0);
signextend_s <= (Others => Imm_p(31));


----------------------------------------------------
-- A 2:1 multiplexer connected to the ALU input B
----------------------------------------------------
with ALUSrcB_p select
	ALUin2_s <=
		B_p			when '0',
		signextend_s & Imm_p			when others;


ALU_Control: process (ALUOp_p, opcode_s, funct_s, ALUSrcB_p, ALUin1_s, ALUin2_s)

begin

	case ALUOp_p is
	
		---------------------------------------------------------------------
 	  	---------------          I-type instructions          ---------------
 	  	---------------------------------------------------------------------
 	  	
		when '0' =>
			case opcode_s is		
			
				when "110111" =>
					-- Load doubleword LD
					ALUout_s <= conv_std_logic_vector(unsigned(ALUin1_s) + unsigned(ALUin2_s), 64);
					
				when "111111" =>
					-- Store doubleword SD
					ALUout_s <= conv_std_logic_vector(unsigned(ALUin1_s) + unsigned(ALUin2_s), 64);
					
				when "011000" =>
					-- Doubleword add immediate DADDI
					ALUout_s <= conv_std_logic_vector(signed(ALUin1_s)+ signed(ALUin2_s), 64);
					
				when "011001" =>
					-- Doubleword add immediate unsigned DADDIU
					ALUout_s <= conv_std_logic_vector(unsigned(ALUin1_s)+ unsigned(ALUin2_s), 64);
					
				when "011010" =>
					-- Doubleword substract immediate DSUBI
					ALUout_s <= conv_std_logic_vector(signed(ALUin1_s)- signed(ALUin2_s), 64);
					
				when "011011" =>
					-- Doubleword substract immediate unsigned DSUBIU
					ALUout_s <= conv_std_logic_vector(unsigned(ALUin1_s)- unsigned(ALUin2_s), 64);
					
				when "001100" =>
					-- And immediate ANDI
					ALUout_s <= ALUin1_s AND ALUin2_s;
					
				when "001101" =>
					-- Or immediate ORI
					ALUout_s <= ALUin1_s OR ALUin2_s;
					
				when "001110" =>
					-- Exclusive or immediate XORI
					ALUout_s <= ALUin1_s XOR ALUin2_s;
					
				when others =>
					ALUout_s <= (Others => '0');
										
			end case;
				
				
		---------------------------------------------------------------------
 	  	---------------          R-type instructions          ---------------
 	  	---------------------------------------------------------------------
 	  	
  	 	when others =>
			case funct_s is
			
				when "000000" =>
					-- No operation NOP
					ALUout_s <= (Others => '0');
					
				when "101100" =>
					-- Doubleword add DADD
					ALUout_s <= conv_std_logic_vector(signed(ALUin1_s)+ signed(ALUin2_s), 64);
					
				when "101101" =>
					-- Doubleword add unsigned DADDU
					ALUout_s <= conv_std_logic_vector(unsigned(ALUin1_s)+ unsigned(ALUin2_s), 64);
					
				when "101110" =>
					-- Doubleword substract DSUB
					ALUout_s <= conv_std_logic_vector(signed(ALUin1_s)- signed(ALUin2_s), 64);
					
				when "101111" =>
					-- Doubleword substract unsigned DSUBU
					ALUout_s <= conv_std_logic_vector(unsigned(ALUin1_s)- unsigned(ALUin2_s), 64);
					
				when "100100" =>
					-- And AND
					ALUout_s <= ALUin1_s AND ALUin2_s;
					
				when "100101" =>
					-- Or OR
					ALUout_s <= ALUin1_s OR ALUin2_s;
				
				when "100110" =>
					-- Exclusive or XOR
					ALUout_s <= ALUin1_s XOR ALUin2_s;
					
				when "111000" =>
					-- Doubleword shift left logical DSLL
					ALUout_s <= to_stdlogicvector(to_bitvector(ALUin1_s) SLL conv_integer(unsigned(ALUin2_s)));
					
				when "111010" =>
					-- Doubleword shift right logical DSRL
					ALUout_s <= to_stdlogicvector(to_bitvector(ALUin1_s) SRL conv_integer(unsigned(ALUin2_s)));
					
				when others =>
					ALUout_s <= (Others => '0');
										
			end case;

	end case;

end process;


end implementation;

