------------------------------------------------------------------------
-- Filename: MIPS64.vhd
-- Title: All components connected
-- Author: Simon Foucher
-- Date: February 2009
------------------------------------------------------------------------

library ieee;
use	ieee.std_logic_1164.all;
use	ieee.std_logic_arith.all;



ENTITY MIPS64 IS
PORT (
	CLOCK : in std_logic;
	RESET : in std_logic);
	
END MIPS64;


ARCHITECTURE logic OF MIPS64 IS

COMPONENT PIPE_CONTROLLER
      PORT(
		INSTRUCTION : in std_logic_vector(31 downto 0);
		HAZARD 		: in std_logic;
		RegDst 		: out std_logic;	-- Controls the write register access
		ALUOp 		: out std_logic; 	-- Controls dependence on funct field
		ALUSrcB 	: out std_logic; 	-- Controls where second ALU operand comes from
		MemRead 	: out std_logic; 	-- Fills memory data output when aserted
		MemWrite 	: out std_logic; 	-- Writes data input to memory when asserted
		RegWrite 	: out std_logic; 	-- Sends a write command
		MemtoReg	: out std_logic;	-- Selects ALU/Memory data
		HAZARD_CTRL : out std_logic); 	-- Notifies PC and Pipeline registers of hazard
END COMPONENT;

COMPONENT hazard_detection
port (
	INST_FROM_IFID	: in std_logic_vector(31 downto 0); -- from the output of the IF/ID stage
	INST_FROM_IDEX	: in std_logic_vector(31 downto 0); -- from the output of the ID/EX stage
	INST_FROM_EXMEM	: in std_logic_vector(31 downto 0); -- from the output of the EX/MEM stage
	INST_FROM_MEMWB	: in std_logic_vector(31 downto 0); -- from the output of the MEM/WB stage
	hazard_p		: out std_logic); -- stop the PC from incrementing and insert a bubble in the pipeline if '1', else keep incrementing the PC
END COMPONENT;

COMPONENT IF_STAGE
   PORT(
		CLK					: in std_logic;
		RST         		: in std_logic;
		HAZARD_MODE			: in std_logic;
		INSTRUCTION			: out std_logic_vector(31 downto 0));
END COMPONENT;

COMPONENT REG_IF_ID
      PORT(
		CLK				: in std_logic;
		RST 			: in std_logic;
		HAZARD			: in std_logic;
		INSTRUCTION_IN 	: in std_logic_vector(31 downto 0);
		INSTRUCTION_OUT : out std_logic_vector(31 downto 0));
END COMPONENT;

COMPONENT ID_WB_STAGE
      PORT(
		RegDst		: in std_logic;	-- Controls the write register access (STAGE ID)
		MemtoReg	: in std_logic;	-- Selects ALU/Memory data (STAGE WB)
		RegWrite	: in std_logic;	-- Sends a write command (STAGE WB)
		Reset		: in std_logic;	-- Resets the system and formats registers
		Instruction	: in std_logic_vector(25 downto 0);
		FromMemDataReg	: in std_logic_vector(63 downto 0);		-- Data from Memory
		FromALUDataReg	: in std_logic_vector(63 downto 0);		-- Data from ALU
		WriteReg_IN		: in std_logic_vector(4 downto 0);		-- Delayed to WB stage
		ReadDataA	: out std_logic_vector(63 downto 0);
		ReadDataB	: out std_logic_vector(63 downto 0);
		SignExtended: out std_logic_vector(31 downto 0);
		WriteReg_OUT: out std_logic_vector(4 downto 0));		-- Decoded in ID then passed on to be delayed to WB
END COMPONENT;

COMPONENT REG_ID_EX
      PORT(
		CLK	: in std_logic;
		RST : in std_logic;
		HAZARD : in std_logic;
		Data_RegA_IN	: in std_logic_vector(63 downto 0);
		Data_RegB_IN	: in std_logic_vector(63 downto 0);
		Extended_IN		: in std_logic_vector(31 downto 0);
		WriteReg_IN		: in std_logic_vector(4 downto 0); 
		Data_RegA_OUT	: out std_logic_vector(63 downto 0);
		Data_RegB_OUT	: out std_logic_vector(63 downto 0);
		Extended_OUT	: out std_logic_vector(31 downto 0);-- To be passed on to WB stage
		WriteReg_OUT	: out std_logic_vector(4 downto 0);	-- To Be passed on to WB stage
		ALU_OP_IN		: in std_logic;
		ALU_Source_IN	: in std_logic;
		ALU_OP_OUT		: out std_logic;
		ALU_Source_OUT	: out std_logic;
	    MemRead_IN     : in std_logic;
  	    MemWrite_IN    : in std_logic;
  	    MemRead_OUT    : out std_logic;
  	    MemWrite_OUT   : out std_logic;
		MemtoReg_IN		: in std_logic;	
		RegWrite_IN		: in std_logic;	
		MemtoReg_OUT	: out std_logic;	
		RegWrite_OUT	: out std_logic;	
		INSTRUCTION_IN : in std_logic_vector(31 downto 0);
		INSTRUCTION_OUT : out std_logic_vector(31 downto 0));
END COMPONENT;

COMPONENT EX_STAGE
port (
	ALUOp_p	  	: in std_logic;	-- the ALU performs an operation based on the opcode field of the instruction if '0' and based on the funct field of the instruction if '1'	
	ALUSrcB_p	: in std_logic; -- select the contents of register B if '0' and the sign extended immediate if '1'
	-- Datapath Inputs
	Instr_p		: in std_logic_vector(31 downto 0);
	Imm_p 		: in std_logic_vector(31 downto 0);	
	A_p		: in std_logic_vector(63 downto 0);
	B_p		: in std_logic_vector(63 downto 0);
	-- Outputs
	ALUOuput_p 		: out std_logic_vector(63 downto 0));
end COMPONENT;

COMPONENT REG_EX_MEM
      PORT(
    	CLK	: in std_logic;
		RST : in std_logic;
		-- DATA PORTS
		ALUResult_IN	: in std_logic_vector(63 downto 0);
		ReadRegB_IN		: in std_logic_vector(63 downto 0); 
		WriteReg_IN		: in std_logic_vector(4 downto 0); 

		ALUResult_OUT	: out std_logic_vector(63 downto 0);
		ReadRegB_OUT	: out std_logic_vector(63 downto 0); 
		WriteReg_OUT	: out std_logic_vector(4 downto 0);	-- To Be passed on to WB stage
		-- CONTROL PORTS
		-- For MEM STAGE
	    MemRead_IN     : in std_logic;
  	    MemWrite_IN    : in std_logic;
  	    MemRead_OUT    : out std_logic;
  	    MemWrite_OUT   : out std_logic;	
		-- For WB STAGE
		MemtoReg_IN		: in std_logic;	
		RegWrite_IN		: in std_logic;	
		MemtoReg_OUT	: out std_logic;	
		RegWrite_OUT	: out std_logic;
		INSTRUCTION_IN : in std_logic_vector(31 downto 0);
		INSTRUCTION_OUT : out std_logic_vector(31 downto 0));	
END COMPONENT;

COMPONENT MEM_STAGE
      PORT(
         ADDRESS     : in  std_logic_vector(63 downto 0);
 	      MemRead     : in std_logic;
  	      MemWrite    : in std_logic;
  	      RST   		: in std_logic;
 	      INData	   : in std_logic_vector (63 downto 0);
 	      OUTData	  : out std_logic_vector (63 downto 0));     
END COMPONENT;

COMPONENT REG_MEM_WB
      PORT(     
		CLK	: in std_logic;
		RST : in std_logic;	
		-- DATA PORTS
		MemData_IN		: in std_logic_vector(63 downto 0); 
		ALUResult_IN	: in std_logic_vector(63 downto 0);
		WriteReg_IN		: in std_logic_vector(4 downto 0); 
		
		MemData_OUT		: out std_logic_vector(63 downto 0); 
		ALUResult_OUT	: out std_logic_vector(63 downto 0);
		WriteReg_OUT	: out std_logic_vector(4 downto 0);
		-- CONTROL PORTS
		-- For WB STAGE
		MemtoReg_IN		: in std_logic;	
		RegWrite_IN		: in std_logic;	
		MemtoReg_OUT	: out std_logic;	
		RegWrite_OUT	: out std_logic;
		INSTRUCTION_IN : in std_logic_vector(31 downto 0);
		INSTRUCTION_OUT : out std_logic_vector(31 downto 0));	
END COMPONENT;



-- Signals are named by 'Signal name'_'origin'
-- Signals from Controller
SIGNAL RegDst_Ctrl 		: std_logic;	-- Controls the write register access
SIGNAL ALUOp_Ctrl 		: std_logic; 	-- Controls dependence on funct field
SIGNAL ALUSrcB_Ctrl 	: std_logic; 	-- Controls where second ALU operand comes from
SIGNAL MemRead_Ctrl 	: std_logic; 	-- Fills memory data output when aserted
SIGNAL MemWrite_Ctrl 	: std_logic; 	-- Writes data input to memory when asserted
SIGNAL RegWrite_Ctrl 	: std_logic; 	-- Sends a write command
SIGNAL MemtoReg_Ctrl	: std_logic;	-- Selects ALU/Memory data
SIGNAL HAZARD_Ctrl 		: std_logic; 	-- Notifies PC and Pipeline registers of hazard

-- Signals from Hazard Detection Unit
SIGNAL hazard_HDU		: std_logic; -- stop the PC from incrementing and insert a bubble in the pipeline if '1', else keep incrementing the PC

-- Signals from IF stage (S = Stage)
SIGNAL	instruction_IFS	: std_logic_vector (31 downto 0);

-- Signals from IF/ID register (R = Register)
SIGNAL INSTRUCTION_IFIDR : std_logic_vector(31 downto 0);

-- Signals from ID stage (S = Stage)
SIGNAL ReadDataA_IDS	: std_logic_vector(63 downto 0);
SIGNAL ReadDataB_IDS	: std_logic_vector(63 downto 0);
SIGNAL SignExtended_IDS : std_logic_vector(31 downto 0);
SIGNAL WriteReg_OUT_IDS : std_logic_vector(4 downto 0);		-- Decoded in ID then passed on to be delayed to WB

-- Signals from ID/EX register (R = Register)
SIGNAL Data_RegA_IDEXR	: std_logic_vector(63 downto 0);
SIGNAL Data_RegB_IDEXR	: std_logic_vector(63 downto 0);
SIGNAL Extended_IDEXR	: std_logic_vector(31 downto 0);-- To be passed on to WB stage
SIGNAL WriteReg_IDEXR	: std_logic_vector(4 downto 0);	-- To Be passed on to WB stage
SIGNAL ALU_OP_IDEXR		: std_logic;
SIGNAL ALU_Source_IDEXR	: std_logic;
SIGNAL MemRead_IDEXR    : std_logic;
SIGNAL MemWrite_IDEXR   : std_logic;
SIGNAL MemtoReg_IDEXR	: std_logic;	
SIGNAL RegWrite_IDEXR	: std_logic;
SIGNAL INSTRUCTION_IDEXR : std_logic_vector(31 downto 0);
	

-- Signals from EX stage (S = Stage)
SIGNAL ALUOuput_EXS		: std_logic_vector(63 downto 0);

-- Signals from EX/MEM register (R = Register)
SIGNAL ALUResult_EXMEMR	: std_logic_vector(63 downto 0);
SIGNAL ReadRegB_EXMEMR	: std_logic_vector(63 downto 0); 
SIGNAL WriteReg_EXMEMR	: std_logic_vector(4 downto 0);	-- To Be passed on to WB stage
SIGNAL MemRead_EXMEMR   : std_logic;
SIGNAL MemWrite_EXMEMR  : std_logic;	
SIGNAL MemtoReg_EXMEMR	: std_logic;	
SIGNAL RegWrite_EXMEMR	: std_logic;
SIGNAL INSTRUCTION_EXMEMR : std_logic_vector(31 downto 0);
	

-- Signals from MEM stage (S = Stage)
SIGNAL OUTData_MEMS	  : std_logic_vector (63 downto 0);     

-- Signals from MEM/WB register (R = Register)
SIGNAL MemData_MEMWBR		: std_logic_vector(63 downto 0); 
SIGNAL ALUResult_MEMWBR		: std_logic_vector(63 downto 0);
SIGNAL WriteReg_MEMWBR		: std_logic_vector(4 downto 0);
SIGNAL MemtoReg_MEMWBR		: std_logic;	
SIGNAL RegWrite_MEMWBR		: std_logic;
SIGNAL INSTRUCTION_MEMWBR 	: std_logic_vector(31 downto 0);
	



BEGIN


CTRL: PIPE_CONTROLLER PORT MAP(
		-- Inputs
		INSTRUCTION => INSTRUCTION_IFIDR,
		HAZARD 		=> hazard_HDU,		-- Receives a signal when a Hazard is detected
		-- Outputs
		RegDst 		=> RegDst_Ctrl,		-- Controls the write register access
		ALUOp 		=> ALUOp_Ctrl,		-- Controls dependence on funct field
		ALUSrcB 	=> ALUSrcB_Ctrl,	-- Controls where second ALU operand comes from
		MemRead 	=> MemRead_Ctrl,	-- Fills memory data output when aserted
		MemWrite 	=> MemWrite_Ctrl,	-- Writes data input to memory when asserted
		RegWrite 	=> RegWrite_Ctrl,	-- Sends a write command
		MemtoReg	=> MemtoReg_Ctrl,	-- Selects ALU/Memory data
		HAZARD_CTRL => HAZARD_Ctrl		-- Notifies PC and Pipeline registers of hazard
		); 	

HDU : hazard_detection PORT MAP (
	INST_FROM_IFID	=> INSTRUCTION_IFIDR,	-- from the output of the IF/ID stage
	INST_FROM_IDEX	=> INSTRUCTION_IDEXR,	-- from the output of the ID/EX stage
	INST_FROM_EXMEM	=> INSTRUCTION_EXMEMR,	-- from the output of the EX/MEM stage
	INST_FROM_MEMWB	=> INSTRUCTION_MEMWBR,	-- from the output of the MEM/WB stage
	hazard_p		=> hazard_HDU -- stop the PC from incrementing and insert a bubble in the pipeline if '1', else keep incrementing the PC
	);

Stage_Fetch : IF_STAGE PORT MAP(
		CLK					=> CLOCK,
		RST					=> RESET,				
		HAZARD_MODE			=> HAZARD_Ctrl,			-- Hazard detected 
		INSTRUCTION			=> instruction_IFS		-- Instruction output (_IFS = from IFS)
		);

Register_IFID : REG_IF_ID PORT MAP(
		CLK				=> CLOCK,
		RST 			=> RESET,
		HAZARD			=> HAZARD_Ctrl,				-- Hazard Signaled from the controller
		-- Instruction carry forward for hazard detection
		INSTRUCTION_IN 	=> instruction_IFS,			-- Instruction from IF Stage
		INSTRUCTION_OUT => INSTRUCTION_IFIDR		-- Instruction output (_IFIDR = from IF/ID Register)
		);
		
Stage_ID_WB : ID_WB_STAGE PORT MAP(
		Reset			=> RESET,					-- Resets the system and formats registers
		-- Control Signals for ID Stage
		RegDst			=> RegDst_Ctrl,				-- Controls the write register access (STAGE ID)
		-- Control Signals for WB Stage
		MemtoReg		=> MemtoReg_MEMWBR,			-- Selects ALU/Memory data (MEMWBR = From MEM/WB Register)
		RegWrite		=> RegWrite_MEMWBR,			-- Sends a write command (MEMWBR = From MEM/WB Register)
		-- DATA ports for ID Stage
		Instruction		=> INSTRUCTION_IFIDR(25 downto 0),  -- Rest of the instruction (op code) is passed to controlle				-- DATA ports for ID Stage
		-- DATA ports for WB Stage (From MEM/WB Register)
		FromMemDataReg	=> MemData_MEMWBR,			-- Data from Memory
		FromALUDataReg	=> ALUResult_MEMWBR,		-- Data from ALU
		WriteReg_IN		=> WriteReg_MEMWBR,			-- Delayed to WB stage
		-- Data Output
		ReadDataA		=> ReadDataA_IDS,			-- Data from Register A
		ReadDataB		=> ReadDataB_IDS,			-- Data from Register B
		SignExtended	=> SignExtended_IDS,		-- Sign Extended Immediate
		WriteReg_OUT	=> WriteReg_OUT_IDS			-- Decoded in ID then passed on to be delayed to WB	
		);	
		
Register_IDEX : COMPONENT REG_ID_EX PORT MAP(
		CLK				=> CLOCK,
		RST 			=> RESET,
		HAZARD 			=> HAZARD_Ctrl,
		-- Data Input Ports (From ID Stage)
		Data_RegA_IN	=> ReadDataA_IDS,			-- Data from Register A
		Data_RegB_IN	=> ReadDataB_IDS,			-- Data from Register B
		Extended_IN		=> SignExtended_IDS,		-- Sign Extended Immediate
		WriteReg_IN		=> WriteReg_OUT_IDS,		-- Decoded in ID then passed on to be delayed to WB
		-- Data Input Ports (_IDEXR = from ED/EX Regiater)
		Data_RegA_OUT	=> Data_RegA_IDEXR,			-- Data from Register A
		Data_RegB_OUT	=> Data_RegB_IDEXR,			-- Data from Register B
		Extended_OUT	=> Extended_IDEXR,			-- To be passed on to WB stage
		WriteReg_OUT	=> WriteReg_IDEXR,			-- To Be passed on to WB stage
		-- CONTROL Signals
		-- For EX STAGE
		ALU_OP_IN		=> ALUOp_Ctrl,				-- In ID/EX From the controller
		ALU_Source_IN	=> ALUSrcB_Ctrl,			-- In ID/EX the controller
		ALU_OP_OUT		=> ALU_OP_IDEXR,			-- Out From ID EX Register
		ALU_Source_OUT	=> ALU_Source_IDEXR,		-- Out From ID EX Register
		-- For MEM STAGE
	    MemRead_IN     	=> MemRead_Ctrl,			-- In ID/EX From the controller
  	    MemWrite_IN    	=> MemWrite_Ctrl,			-- In ID/EX From the controller
  	    MemRead_OUT    	=> MemRead_IDEXR,			-- Out From ID EX Register
  	    MemWrite_OUT   	=> MemWrite_IDEXR,			-- Out From ID EX Register
   		-- For WB STAGE
		MemtoReg_IN		=> MemtoReg_Ctrl,			-- In ID/EX From the controller
		RegWrite_IN		=> RegWrite_Ctrl,			-- In ID/EX From the controller
		MemtoReg_OUT	=> MemtoReg_IDEXR,			-- Out From ID EX Register
		RegWrite_OUT	=> RegWrite_IDEXR,			-- FOut From ID EX Register
		-- Instruction carry forward for hazard detection
		INSTRUCTION_IN 	=> INSTRUCTION_IFIDR,		-- In In ID/EX From IF/ID Register
		INSTRUCTION_OUT => INSTRUCTION_IDEXR		-- Out from ID/EX Register
		);


Stage_EX : EX_STAGE PORT MAP (
		-- Control Signals
		ALUOp_p	  	=> ALU_OP_IDEXR, 				-- In ALU from ID/EX Register				
		ALUSrcB_p	=> ALU_Source_IDEXR,			-- In ALU from ID/EX Register
		-- Data Ports	
		Instr_p		=> INSTRUCTION_IDEXR,			-- In ALU from ID/EX Register	
		Imm_p 		=> Extended_IDEXR,				-- Immediate Fiels Sign Extended In ALU from ID/EX Register	
		A_p			=> Data_RegA_IDEXR,				-- Data from RegA In ALU from ID/EX Register	
		B_p			=> Data_RegB_IDEXR,				-- Data from RegB In ALU from ID/EX Register	
		ALUOuput_p 	=> ALUOuput_EXS					-- Output of the ALU (_EXS = From EX Stage)
		);
		
Register_EX_MEM : REG_EX_MEM PORT MAP(
		CLK				=> CLOCK,
		RST 			=> RESET,
		-- DATA PORTS (From Stage EX)
		ALUResult_IN	=> ALUOuput_EXS,			-- ALU output into EX/MEM Register
		ALUResult_OUT	=> ALUResult_EXMEMR,		-- ALU Output from EX/MEM Register

		-- DATA PORTS (From Register ID/EXE)
		ReadRegB_IN		=> Data_RegB_IDEXR,			-- Data From Register B into EX/MEM Register (From ID/EX Register)
		WriteReg_IN		=> WriteReg_IDEXR,			-- Write Register carry forward into EX/MEM Register (From ID/EX Register)

		ReadRegB_OUT	=> ReadRegB_EXMEMR,			-- Data From Register B (From ID/EX Register)
		WriteReg_OUT	=> WriteReg_EXMEMR,			-- Write Register carry forward (From ID/EX Register)
		
		-- CONTROL PORTS
		-- For MEM STAGE
	    MemRead_IN     => MemRead_IDEXR,			-- From ID/EX Register into EX/MEM Register
  	    MemWrite_IN    => MemWrite_IDEXR,			-- From ID/EX Register into EX/MEM Register
  	    MemRead_OUT    => MemRead_EXMEMR,			-- Output From EX/MEM Register
  	    MemWrite_OUT   => MemWrite_EXMEMR,			-- Output From EX/MEM Register
		-- For WB STAGE
		MemtoReg_IN		=> MemtoReg_IDEXR,			-- From ID/EX Register into EX/MEM Register
		RegWrite_IN		=> RegWrite_IDEXR,			-- From ID/EX Register into EX/MEM Register
		MemtoReg_OUT	=> MemtoReg_EXMEMR,			-- Output From EX/MEM Register
		RegWrite_OUT	=> RegWrite_EXMEMR,			-- Output From EX/MEM Register
		-- Instruction carry forward for hazard detection
		INSTRUCTION_IN	=> INSTRUCTION_IDEXR,
		INSTRUCTION_OUT => INSTRUCTION_EXMEMR
		);
	
	
		
Stage_Memory : MEM_STAGE PORT MAP(
  	     RST		=> RESET,
         ADDRESS	=> ALUResult_EXMEMR,			-- Target Address Read from MEM/EX Register
 	     MemRead	=> MemRead_EXMEMR,				-- Memory Read Command Read from MEM/EX Register
  	     MemWrite	=> MemWrite_EXMEMR,				-- Memory Write Command Read from MEM/EX Register
  	     INData		=> ReadRegB_EXMEMR,				-- Data to be written into MEM Read from MEM/EX Register
 	     OUTData	=> OUTData_MEMS					-- Output from Memory (For LD Operations)
 	     );  
 	     
 Register_MEMWB : REG_MEM_WB PORT MAP(     
		CLK				=> CLOCK,
		RST 			=> RESET,
		-- DATA PORTS
		MemData_IN		=> OUTData_MEMS,			-- Data read from memory fed Into MEM/WB Register
		ALUResult_IN	=> ALUResult_EXMEMR,		-- ALU Result Carried forward from EX/MEM Register into MEM/WB Register
		WriteReg_IN		=> WriteReg_EXMEMR,			-- Target Register Carried forward from EX/MEM Register into MEM/WB Register
		
		MemData_OUT		=> MemData_MEMWBR,
		ALUResult_OUT	=> ALUResult_MEMWBR,
		WriteReg_OUT	=> WriteReg_MEMWBR,
		-- CONTROL PORTS
		-- For WB STAGE
		MemtoReg_IN		=> MemtoReg_EXMEMR,
		RegWrite_IN		=> RegWrite_EXMEMR,
		
		MemtoReg_OUT	=> MemtoReg_MEMWBR,
		RegWrite_OUT	=> RegWrite_MEMWBR,
		-- Instruction carry forward for hazard detection
		INSTRUCTION_IN	=> INSTRUCTION_EXMEMR,
		INSTRUCTION_OUT => INSTRUCTION_MEMWBR
		);		     
 	     
 	     
 	     
 	     
		
END logic;




