
--------------------------------------------------------------------------------------------------------------------------
-- Filename: hazard_detection.vhd
-- Title: Implementation of control logic that determines whether a hazard could/will occur
-- Author: Yvan Ton-That
-- Date: February 2009
--------------------------------------------------------------------------------------------------------------------------

--------------------------------------------------------------------------------------------------------------------------
-- Purpose: The hazard detection unit operates during the ID stage so that it can insert a stall between the two 
-- instructions that cause the detected hazard. This is done by changing the instruction in the ID stage into a NOP. 
-- If the instruction in the ID stage us stalled, then the instruction in the IF stage must also be stalled; otherwise, 
-- we would lose the fetched instruction. Preventing the instructions in the IF and ID stages from making progress is 
-- accomplished simply by preventing the PC register and the IF/ID pipeline register from changing. Provided these 
-- registers are preserved, the instruction in the IF stage will continue to be read using the same PC, and the registers 
-- in the ID stage will continue to be read using the same instruction fields in the IF/ID pipeline register. Thus, 
-- before the next instruction (which would cause the hazard) is executed, the previous one will have had sufficient 
-- time to complete and prevent the hazard.
--------------------------------------------------------------------------------------------------------------------------


LIBRARY ieee;
USE	ieee.std_logic_1164.ALL;


ENTITY hazard_detection IS

port (

	-- Inputs
	INST_FROM_IFID	: in std_logic_vector(31 downto 0); -- from the output of the IF/ID stage
	INST_FROM_IDEX	: in std_logic_vector(31 downto 0); -- from the output of the ID/EX stage
	INST_FROM_EXMEM	: in std_logic_vector(31 downto 0); -- from the output of the EX/MEM stage
	INST_FROM_MEMWB	: in std_logic_vector(31 downto 0); -- from the output of the MEM/WB stage

	-- Outputs
	hazard_p	: out std_logic -- stop the PC from incrementing and insert a bubble in the pipeline if '1', else keep incrementing the PC

);

END hazard_detection;




ARCHITECTURE implementation OF hazard_detection IS

BEGIN


hazard_detection_unit : PROCESS ( INST_FROM_IFID, INST_FROM_IDEX, INST_FROM_EXMEM, INST_FROM_MEMWB )

--------------------------------------------------------------------------------------------------------------------------
-- CONSUMER	PRODUCER		STATUS
-- 
-- ALU	    ALU		ok
-- ALU 	    ALUimm		ok
-- ALU 	    Load		ok (same as ALU - ALUimm)
-- ALU 	    Store		no hazard possible
-- 
-- ALUimm 	 ALU		ok
-- ALUimm 	 ALUimm		ok
-- ALUimm 	 Load		ok (same as ALUimm - ALUimm)
-- ALUimm 	 Store		no hazard possible
-- 
-- Load 	   ALU		ok (same as ALUimm - ALU)
-- Load 	   ALUimm		ok (same as ALUimm - ALUimm)
-- Load 	   Load		ok (same as ALUimm - ALUimm)
-- Load 	   Store		no hazard possible
-- 
-- Store 	  ALU		ok (same as ALUimm - ALU)
-- Store 	  ALUimm		ok (same as ALUimm - ALUimm)
-- Store 	  Load		ok (same as ALUimm - ALUimm)
-- Store 	  Store		no hazard possible
--------------------------------------------------------------------------------------------------------------------------

BEGIN


	--------------------------------------------------------------------------------------------------------------------------
	-- Consumer : an ALU instruction
	-- Producer : an ALU instruction
	--------------------------------------------------------------------------------------------------------------------------

	IF INST_FROM_IFID (31 downto 26) = "000000" AND INST_FROM_IDEX (31 downto 26) = "000000" THEN

		IF INST_FROM_IFID (25 downto 21) = INST_FROM_IDEX (15 downto 11) OR INST_FROM_IFID (20 downto 16) = INST_FROM_IDEX (15 downto 11) THEN hazard_p <= '1';
				
		ELSE hazard_p <= '0';
			
		END IF;

	
	ELSIF INST_FROM_IFID (31 downto 26) = "000000" AND INST_FROM_EXMEM (31 downto 26) = "000000" THEN
	
		IF INST_FROM_IFID (25 downto 21) = INST_FROM_EXMEM (15 downto 11) OR INST_FROM_IFID (20 downto 16) = INST_FROM_EXMEM (15 downto 11) THEN hazard_p <= '1';
				
		ELSE hazard_p <= '0';
			
		END IF;


	ELSIF INST_FROM_IFID (31 downto 26) = "000000" AND INST_FROM_MEMWB (31 downto 26) = "000000" THEN

		IF INST_FROM_IFID (25 downto 21) = INST_FROM_MEMWB (15 downto 11) OR INST_FROM_IFID (20 downto 16) = INST_FROM_MEMWB (15 downto 11) THEN hazard_p <= '1';
				
		ELSE hazard_p <= '0';
			
		END IF;


	--------------------------------------------------------------------------------------------------------------------------
	-- Consumer : an ALU instruction
	-- Producer : an ALU immidiate instruction or a load instruction
	--------------------------------------------------------------------------------------------------------------------------

	ELSIF INST_FROM_IFID (31 downto 26) = "000000" AND INST_FROM_IDEX (31 downto 26) /= "000000" AND INST_FROM_IDEX (31 downto 26) /= "111111" THEN

		IF INST_FROM_IFID (25 downto 21) = INST_FROM_IDEX (20 downto 16) OR INST_FROM_IFID (20 downto 16) = INST_FROM_IDEX (20 downto 16) THEN hazard_p <= '1';
				
		ELSE hazard_p <= '0';
			
		END IF;

	
	ELSIF INST_FROM_IFID (31 downto 26) = "000000" AND INST_FROM_EXMEM (31 downto 26) /= "000000" AND INST_FROM_EXMEM (31 downto 26) /= "111111" THEN
	
		IF INST_FROM_IFID (25 downto 21) = INST_FROM_EXMEM (20 downto 16) OR INST_FROM_IFID (20 downto 16) = INST_FROM_EXMEM (20 downto 16) THEN hazard_p <= '1';
				
		ELSE hazard_p <= '0';
			
		END IF;


	ELSIF INST_FROM_IFID (31 downto 26) = "000000" AND INST_FROM_MEMWB (31 downto 26) /= "000000" AND INST_FROM_MEMWB (31 downto 26) /= "111111" THEN

		IF INST_FROM_IFID (25 downto 21) = INST_FROM_MEMWB (20 downto 16) OR INST_FROM_IFID (20 downto 16) = INST_FROM_MEMWB (20 downto 16) THEN hazard_p <= '1';
				
		ELSE hazard_p <= '0';
			
		END IF;


	--------------------------------------------------------------------------------------------------------------------------
	-- Consumer : an ALU immidiate instruction, a load instruction, or a store instruction
	-- Producer : an ALU instruction
	--------------------------------------------------------------------------------------------------------------------------

	ELSIF INST_FROM_IFID (31 downto 26) /= "000000" AND INST_FROM_IDEX (31 downto 26) = "000000" THEN

		IF INST_FROM_IFID (25 downto 21) = INST_FROM_IDEX (15 downto 11) THEN hazard_p <= '1';
				
		ELSE hazard_p <= '0';
			
		END IF;

	
	ELSIF INST_FROM_IFID (31 downto 26) /= "000000" AND INST_FROM_EXMEM (31 downto 26) = "000000" THEN
	
		IF INST_FROM_IFID (25 downto 21) = INST_FROM_EXMEM (15 downto 11) THEN hazard_p <= '1';
				
		ELSE hazard_p <= '0';
			
		END IF;


	ELSIF INST_FROM_IFID (31 downto 26) /= "000000" AND INST_FROM_MEMWB (31 downto 26) = "000000" THEN

		IF INST_FROM_IFID (25 downto 21) = INST_FROM_MEMWB (15 downto 11) THEN hazard_p <= '1';
				
		ELSE hazard_p <= '0';
			
		END IF;


	--------------------------------------------------------------------------------------------------------------------------
	-- Consumer : an ALU immediate instruction, a load instruction, or a store instruction
	-- Producer : an ALU immediate instruction or a load instruction
	--------------------------------------------------------------------------------------------------------------------------

	ELSIF INST_FROM_IFID (31 downto 26) /= "000000" AND INST_FROM_IDEX (31 downto 26) /= "000000" AND INST_FROM_IDEX (31 downto 26) /= "111111" THEN

		IF INST_FROM_IFID (25 downto 21) = INST_FROM_IDEX (20 downto 16) THEN hazard_p <= '1';
				
		ELSE hazard_p <= '0';
			
		END IF;

	
	ELSIF INST_FROM_IFID (31 downto 26) /= "000000" AND INST_FROM_EXMEM (31 downto 26) /= "000000"  AND INST_FROM_EXMEM (31 downto 26) /= "111111" THEN
	
		IF INST_FROM_IFID (25 downto 21) = INST_FROM_EXMEM (20 downto 16) THEN hazard_p <= '1';
				
		ELSE hazard_p <= '0';
			
		END IF;


	ELSIF INST_FROM_IFID (31 downto 26) /= "000000" AND INST_FROM_MEMWB (31 downto 26) /= "000000" AND INST_FROM_MEMWB (31 downto 26) /= "111111" THEN

		IF INST_FROM_IFID (25 downto 21) = INST_FROM_MEMWB (20 downto 16) THEN hazard_p <= '1';
				
		ELSE hazard_p <= '0';
			
		END IF;
		
		
	END IF;


END PROCESS;


END implementation;


